/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#define NBIN 5
/*#define DEBUG 1*/
/**********************************************************
 * Usage:
 *    oirChopPhotometry "AFiles" "BFiles" maskFile outFile -nDrop nDrop -dSky dSky
 *
 *    inFiles is a blank-separated list of input files
 *
 *    Output is a IMAGING_DATA table in a fits file with
 *    12-rows, each with only one DATA array (sum of channels 1 and 2)
 *
 *    1:  Total Flux A-side (ADUs/s/channel)
 *    2:  Total Flux B-side 
 *    3:  Total Flux Geometric Mean AB
 *    4:  Masked Flux A-side 
 *    5:  Masked Flux B-side 
 *    6:  Masked Flux Geometric Mean AB
 *    7->12 are estimated RMS errors in rows 1->6 resp.
 *
 *    nDrop is the number of frames to drop after a shift from Sky to Target
 ************************************************************* */

/***************************************************************
 * read through all raw files associated with one shutter setting
 * separate target and sky frames and accumulate them in
 * given arrays; divide by number of frames
 * INPUTS:
 *    char **string     array of file names
 *    int    nFiles     number of file names
 *    int    chan1      0-based number of 1st interferometric channel
 *    int    chan2      0-based number of 2nd interferometric channel
 *    int    nDrop      drop this many points after change of chop phase
 *    int    npix1      no. pixels in array in 1st channel
 *    int    npix2      no. pixels in array in 2nd channel
 *    float* target1    array to receive target1 data
 *    float* target2    array to receive target2 data
 *    float* sky1       array to receive target1 data
 *    float* sky2       array to receive target2 data
 *    int  * nTarget    number of target frames
 *    int  * nSky       number of sky frames
 **************************************************************** */

void unChopData(char **fileList, int nFiles, int chan1, int chan2, 
   int nDrop, long npix1, long npix2, float *target1, float *target2, 
   float *sky1, float *sky2, int *nTarget, int *nSky,
   int *nCycle, int *ierr) {

   fitsfile *inFile;
   imaging_data_table_def *idata;
   imaging_data_row *inData;
   int iFile;

   long iRow, nRow;
   int i;
   int TN, SN;
   long iPix;

   char Sky = 'S';
   char Target = 'T';
   char targ, oldTarg;
   int sinceChange;
   float mjd;
   int old;

   *ierr    = 0;
   *nCycle = 0;
   TN = 0;
   SN = 0;
   oldTarg = ' ';

/*
      loop over files, read rows, read target types, add data
      into appropriate accumulators 
*/

   for (iFile =0; iFile <nFiles; iFile++) {
/*
      open file, open table, create table row
*/
      i = fits_open_file(&inFile, fileList[iFile], READONLY, ierr);
      if (*ierr !=0) {
         printf("error %i opening file %s\n", *ierr, fileList[iFile]);
         printf("%s \n",strerror(errno));
         return;
      }
/*
      get observing data to check chopping flag coding
*/
      i = fits_read_key(inFile, TFLOAT, "MJD-OBS", &mjd,0,ierr);
      old = mjd < 52960.;
      if (*ierr !=0) {
         printf("error %i reading mjd file %s\n", 
            *ierr, fileList[iFile]);
         return;
      }
      idata = open_imaging_data_table(inFile, ierr);
      if (*ierr !=0) {
         printf("error %i opening imaging data in file %s\n", 
            *ierr, fileList[iFile]);
         return;
      }
      inData = make_imaging_data_row(idata);
      i = fits_get_num_rows (inFile,  &nRow, ierr);
      /* suppress autoscaling to avoid overflow */
      for (i=0; i<idata->nregion; i++)
         fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, ierr);

/*
         read lines, check target type and decide whether to accumulate
*/
      for (iRow =1; iRow<=nRow; iRow++) {
         i = read_imaging_data_table(inFile, idata, inData, iRow, ierr);
         if (*ierr !=0) {
            printf("error %i reading imaging data line %ld in file %s\n", 
               *ierr, iRow, fileList[iFile]);
            return;
         }
         if (old) targ = inData->targtype[0]; 
         else targ = inData->targtype[1]; 
         if (targ != oldTarg ) {
            sinceChange = 0;
            oldTarg     = targ;
            if ((targ == Target) || (targ == Sky)) (*nCycle)++;
         } else {
            ++sinceChange;
         }
/*
            if sinceChange > nDrop decide what type and accumulate data
*/
         if (sinceChange > nDrop && (targ == Target)) {
            for (iPix=0;iPix<npix1;iPix++) target1[iPix] =
               target1[iPix] + 32768. + (short)inData->data[chan1][iPix];
            for (iPix=0;iPix<npix2;iPix++) target2[iPix] =
               target2[iPix] + 32768. + (short)inData->data[chan2][iPix];
            ++TN;
         } else if (sinceChange > nDrop && (targ == Sky)) {
            for (iPix=0;iPix<npix1;iPix++) sky1[iPix] =
               sky1[iPix] + 32768. + (short)inData->data[chan1][iPix];
            for (iPix=0;iPix<npix2;iPix++) sky2[iPix] =
               sky2[iPix] + 32768. + (short)inData->data[chan2][iPix];
            ++SN;
         }
      } /* read rows in one file */
      
/*
         close file and table, destroy row
*/
      i= close_imaging_data_table(inFile, idata, ierr);
      kill_imaging_data_row(inData);
   } /* read all files in list */
   *nTarget = TN;
   *nSky    = SN;
   (*nCycle) = (*nCycle)/2;
   return;
} /* end of unChopData */

void binChopData(char **fileList, int nFiles, int chan1, int chan2, 
   int nDrop, long npix1, long npix2, int nCycle,
   float *a1[NBIN], float *a2[NBIN], int *ierr) {

   fitsfile *inFile;
   imaging_data_table_def *idata;
   imaging_data_row *inData;
   int iFile;

   long iRow, nRow;
   int i;
   int TN, SN;
   long iPix;

   char Sky = 'S';
   char Target = 'T';
   char targ, oldTarg;
   int sinceChange;
   float mjd;
   int old;

   float *aTarg1, *aTarg2, *aSky1, *aSky2; /* chop accumulators */
   int iBin, nBin;
   int iCycle = 0;
   *ierr = 0;
   TN = 0;
   SN = 0;
   oldTarg = ' ';



/*
      how many cycles per bin?
*/
   nBin = nCycle/NBIN;
/*
      allocate single cycle accumlators
*/
   aTarg1 = (float *)malloc(npix1 * sizeof(float));
   aTarg2 = (float *)malloc(npix2 * sizeof(float));
   aSky1  = (float *)malloc(npix1 * sizeof(float));
   aSky2  = (float *)malloc(npix2 * sizeof(float));
/*
      allocate bin accumulators
/*
   for (i=0; i<NBIN; i++) {
      a1[i] = (float *)malloc(npix1 *sizeof(float));
      a2[i] = (float *)malloc(npix2 *sizeof(float));
      memset(a1, 0, npix1*sizeof(float));
      memset(a2, 0, npix2*sizeof(float));
   }
/*
      loop over files, read rows, read target types, add data
      into appropriate accumulators 
*/

   for (iFile =0; iFile <nFiles; iFile++) {
/*
      open file, open table, create table row
*/
      i = fits_open_file(&inFile, fileList[iFile], READONLY, ierr);
      if (*ierr !=0) {
         printf("error %i opening file %s\n", *ierr, fileList[iFile]);
         printf("%s \n",strerror(errno));
         return;
      }
      i = fits_read_key(inFile, TFLOAT, "MJD-OBS", &mjd,0,ierr);
      old = mjd < 52960.;
      if (*ierr !=0) {
         printf("error %i reading mjd file %s\n", 
            *ierr, fileList[iFile]);
         return;
      }
      idata = open_imaging_data_table(inFile, ierr);
      if (*ierr !=0) {
         printf("error %i opening imaging data in file %s\n", 
            *ierr, fileList[iFile]);
         return;
      }
      inData = make_imaging_data_row(idata);
      i = fits_get_num_rows (inFile,  &nRow, ierr);
      /* suppress autoscaling to avoid overflow */
      for (i=0; i<idata->nregion; i++)
         fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, ierr);

/*
         read lines, check target type and decide whether to accumulate
*/
      for (iRow =1; iRow<=nRow; iRow++) {
         i = read_imaging_data_table(inFile, idata, inData, iRow, ierr);
         if (*ierr !=0) {
            printf("error %i reading imaging data line %ld in file %s\n", 
               *ierr, iRow, fileList[iFile]);
            return;
         }
         if (old) targ = inData->targtype[0]; 
         else targ = inData->targtype[1]; 
         if (targ != oldTarg ) {
            sinceChange = 0;
            oldTarg     = targ;
            if (targ == Target) {
               TN = 0; 
               memset(aTarg1, 0, npix1*sizeof(float));
               memset(aTarg2, 0, npix2*sizeof(float));
            } else if (targ == Sky) {
/*
       normalize accumulated target and sky
*/
               if (TN > 0) {
                  for (iPix=0;iPix<npix1;iPix++) aTarg1[iPix] =
                     aTarg1[iPix]/TN;
                  for (iPix=0;iPix<npix2;iPix++) aTarg2[iPix] =
                     aTarg2[iPix]/TN;
               }
               if(SN > 0) {
                  for (iPix=0;iPix<npix1;iPix++) aSky1[iPix] =
                     aSky1[iPix]/SN;
                  for (iPix=0;iPix<npix2;iPix++) aSky2[iPix] =
                     aSky2[iPix]/SN;
/*
                 if previous targ exists, accumulate difference
                 into bin
*/
                 iBin = iCycle/nBin;
                 if ((TN > 0) && (iBin < NBIN)) {
                     for (iPix=0;iPix<npix1;iPix++) a1[iBin][iPix] =
                        a1[iBin][iPix] + aTarg1[iPix] - aSky1[iPix];
                     for (iPix=0;iPix<npix1;iPix++) a2[iBin][iPix] =
                        a2[iBin][iPix] + aTarg2[iPix] - aSky2[iPix];
                  }
                  ++iCycle;
               }
/*
                     clear sky accumulators
*/
               SN = 0;
               memset(aSky1, 0, npix1*sizeof(float));
               memset(aSky2, 0, npix2*sizeof(float));
            }
         } else {
            ++sinceChange;
         }
/*
            if sinceChange > nDrop decide what type and accumulate data
*/
         if (sinceChange > nDrop && (targ == Target)) {
            for (iPix=0;iPix<npix1;iPix++) aTarg1[iPix] =
               aTarg1[iPix] + 32768. + (short)inData->data[chan1][iPix];
            for (iPix=0;iPix<npix2;iPix++) aTarg2[iPix] =
               aTarg2[iPix] + 32768. + (short)inData->data[chan2][iPix];
            ++TN;
         } else if (sinceChange > nDrop && (targ == Sky)) {
            for (iPix=0;iPix<npix1;iPix++) aSky1[iPix] =
               aSky1[iPix] + 32768. + (short)inData->data[chan1][iPix];
            for (iPix=0;iPix<npix2;iPix++) aSky2[iPix] =
               aSky2[iPix] + 32768. + (short)inData->data[chan2][iPix];
            ++SN;
         }
      } /* read rows in one file */
      
/*
         close file and table, destroy row
*/
      i= close_imaging_data_table(inFile, idata, ierr);
      kill_imaging_data_row(inData);
   } /* read all files in list */
/*
        normalize bins
*/
   for (iBin =0; iBin<NBIN; iBin++) {
      for (iPix=0;iPix<npix1;iPix++) a1[iBin][iPix] =
         a1[iBin][iPix]/nBin;
      for (iPix=0;iPix<npix1;iPix++) a2[iBin][iPix] =
         a2[iBin][iPix]/nBin;
   } /* iBin */
   free(aTarg1);
   free(aTarg2);
   free(aSky1);
   free(aSky2);
   return;
} /* end of binChopData */

void normalizeExposure (float *data, long nPix, int nFrames, float dit) {
   float norm;
   long iPix;

   if (nFrames <= 0) return;
   norm = 1./(nFrames * dit);
   for (iPix =0; iPix<nPix; iPix++) data[iPix] = data[iPix]*norm;
   return;
}

void deSky (float *data, int nx, int ny, int y1, int y2) {
   int ix, iy;
   float dy, x1, x2;
   float *topSky, *bottomSky;
   float *yLine;
   int dnsky = 2;
/*
      arrays to hold average of a few rows at top and bottom
*/
   topSky = (float *)malloc(nx * sizeof(float));
   memset(topSky, 0, nx*sizeof(float));
   bottomSky = (float *)malloc(nx * sizeof(float));
   memset(bottomSky, 0, nx*sizeof(float));

/*
      average a few rows
*/
   for (iy =-dnsky; iy<=dnsky; iy++) for (ix=0;ix<nx;ix++) {
      topSky[ix]    = topSky[ix] + data[ix + (iy+y2)*(long)nx];
      bottomSky[ix] = bottomSky[ix] + data[ix + (iy+y1)*(long)nx];
   }
   for (ix=0;ix<nx;ix++) {
      topSky[ix]    = topSky[ix]/(2*dnsky+1);
      bottomSky[ix] = bottomSky[ix]/(2*dnsky+1);
   }
   dy = y2 - y1;

/*
      linear fit and subtraction
*/
   for (iy=0; iy<ny; iy++) {
      yLine = data+iy*(long)nx;
      x1 = ((float)iy - y1)/dy;
      x2 = ((float)iy - y2)/dy;
      for (ix=0; ix<nx;ix++) yLine[ix] = yLine[ix] - topSky[ix]*x1 + bottomSky[ix]*x2;
   }

/*
      free arrays
*/
   free(topSky);
   free(bottomSky);
}

int main (int argc, char** argv) {

   char opt1[12];
   char gris[12];
   char shutter[12];
   float dit;

   char **fileList;
   int nFiles;
   int iShutter;
   int i,ierr;
   int nx, ny;
   long npix;
   long iPix;
   int  ix, iy;
   float xx, mm;
   long  ii;
   int iKey;
   int one = 1;
   int nCycle;

   int yTop = 34;
   int yBottom = 0;
   int chan1, chan2;
   int iarg;

   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *outFile = NULL;

#ifdef DEBUG
   fitsfile *debugFile = NULL;
   imaging_fdata_table_def *debugTable;
   imaging_fdata_row       *debugRow;
#endif

   imaging_fdata_table_def *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *outData;
   float *maskImages[12];
   int nxList[12], nyList[12];
   int nRegion;

   char *Ashutter = "AOPEN";
   char *Bshutter = "BOPEN";
   int  nShutter;
   char *highSens = "HIGH_SENS";

   float *ATarget1, *ATarget2; 
   float *ASky1, *ASky2;
   float *BTarget1, *BTarget2; 
   float *BSky1, *BSky2;
   float *ARms1, *ARms2, *BRms1, *BRms2;

   float *ABin1[NBIN], *ABin2[NBIN], *BBin1[NBIN], *BBin2[NBIN];
   float Ascratch, Bscratch, ABscratch;
   int    ATargetN, ASkyN, BTargetN, BSkyN;
   int ANRms = 1;
   int BNRms = 1;

   float *AMask, *BMask, *ABMask;
   float *ATotal, *BTotal, *ABTotal;

   float *ARMask, *BRMask, *ABRMask;
   float *ARTotal, *BRTotal, *ABRTotal;

   int nDrop;
   int dSky;

   ierr = 0;
/*
      see if nDrop, dSky are specified externally
*/
   nDrop = 2;
   dSky  = 0;
   iarg  = 0;
   if (argc > 5) while(iarg < argc ) {
      if(!strcmp("-nDrop", argv[iarg])) sscanf(argv[++iarg], "%i", &nDrop);
      if(!strcmp("-dSky", argv[iarg])) sscanf(argv[++iarg], "%i", &dSky);
      ++iarg;
   }
   printf("nDrop = %i\n", nDrop);
   printf("dSky  = %i\n", dSky);

/*        
       break up file list and print what you find 
       then read and unchop files
*/
   for (iShutter = 1;iShutter<=2; iShutter++) {
      fileList = oirStrSplit(argv[iShutter], &nFiles);

/*
      open first file and get a lot of parameters for iShutter=1
*/
      i = fits_open_file(&inFile, fileList[0], READONLY, &ierr) ;
      if (ierr != 0) {
         printf("cfitsio error %i opening file %s\n", ierr, fileList[0]);
         printf("%s \n",strerror(errno));
         return 1;
      }
      ierr = 0;
      i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS SHUT NAME", 
         shutter, 0, &ierr);
      printf("shutter = %s\n", shutter);
      nShutter = 0;
      if (strncmp(shutter, Ashutter, 5)==0) nShutter=1;
      if (strncmp(shutter, Bshutter, 5)==0) nShutter=2; 
      if (iShutter == 1) {
         i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO DET DIT", &dit, 
            0, &ierr);
         printf("dit = %f\n", dit);
         ierr = 0;
         i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", opt1,0, &ierr);
         printf("beam combiner = %s\n", opt1);

         if (strncmp(opt1, highSens, 9) == 0) { chan1 = 0; chan2 = 1;}
            else {chan1 = 1; chan2 = 2;}
         i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", gris,0,&ierr);
         printf("disperser is = %s\n", gris);
/*
      move to detector section
*/
         DetectorDef = read_imaging_detector_table(inFile, &ierr);
         if (ierr !=0) {
            printf("error %i opening detector table in %s\n", 
               ierr, fileList[0]);
            return 1;
         }
         nRegion = DetectorDef->nregion;
         for (i=0;i<nRegion;i++) {
            nxList[i] = DetectorDef->naxis[i][0];
            nyList[i] = DetectorDef->naxis[i][1];
         }
         nx = DetectorDef->naxis[chan1][0];
         ny = DetectorDef->naxis[chan1][1];

/*
      allocate accumulator arrays
*/
  
         npix = ny * (long)nx;
         ATarget1 = (float *)malloc(npix * sizeof(float));
         memset (ATarget1, 0, npix*sizeof(float));
         BTarget1 = (float *)malloc(npix * sizeof(float));
         memset (BTarget1, 0, npix*sizeof(float));
         ASky1    = (float *)malloc(npix * sizeof(float));
         memset (ASky1, 0, npix*sizeof(float));
         BSky1    = (float *)malloc(npix * sizeof(float));
         memset (BSky1, 0, npix*sizeof(float));
         ATarget2 = (float *)malloc(npix * sizeof(float));
         memset (ATarget2, 0, npix*sizeof(float));
         BTarget2 = (float *)malloc(npix * sizeof(float));
         memset (BTarget2, 0, npix*sizeof(float));
         ASky2    = (float *)malloc(npix * sizeof(float));
         memset (ASky2, 0, npix*sizeof(float));
         BSky2    = (float *)malloc(npix * sizeof(float));
         memset (BSky2, 0, npix*sizeof(float));

         for (i=0; i<NBIN; i++) {
            ABin1[i] = (float *)malloc(npix * sizeof(float));
            memset (ABin1[i], 0, npix*sizeof(float));
            ABin2[i] = (float *)malloc(npix * sizeof(float));
            memset (ABin2[i], 0, npix*sizeof(float));
            BBin1[i] = (float *)malloc(npix * sizeof(float));
            memset (BBin1[i], 0, npix*sizeof(float));
            BBin2[i] = (float *)malloc(npix * sizeof(float));
            memset (BBin2[i], 0, npix*sizeof(float));
         }

         AMask  = (float *)malloc(nx * sizeof(float));
         memset (AMask, 0, nx*sizeof(float));
         BMask  = (float *)malloc(nx * sizeof(float));
         memset (BMask, 0, nx*sizeof(float));
         ABMask = (float *)malloc(nx * sizeof(float));
         memset (ABMask, 0, nx*sizeof(float));
         ATotal  = (float *)malloc(nx * sizeof(float));
         memset (ATotal, 0, nx*sizeof(float));
         BTotal  = (float *)malloc(nx * sizeof(float));
         memset (BTotal, 0, nx*sizeof(float));
         ABTotal = (float *)malloc(nx * sizeof(float));
         memset (ABTotal, 0, nx*sizeof(float));

         ARMask  = (float *)malloc(nx * sizeof(float));
         memset (ARMask, 0, nx*sizeof(float));
         BRMask  = (float *)malloc(nx * sizeof(float));
         memset (BRMask, 0, nx*sizeof(float));
         ABRMask = (float *)malloc(nx * sizeof(float));
         memset (ABRMask, 0, nx*sizeof(float));
         ARTotal  = (float *)malloc(nx * sizeof(float));
         memset (ARTotal, 0, nx*sizeof(float));
         BRTotal  = (float *)malloc(nx * sizeof(float));
         memset (BRTotal, 0, nx*sizeof(float));
         ABRTotal = (float *)malloc(nx * sizeof(float));
         memset (ABRTotal, 0, nx*sizeof(float));

      }  /* iShutter = 1 */
      i = fits_close_file(inFile, &ierr);
/*
          now process the real data
*/
      if (nShutter == 1) {
         printf("Found %i A-photometry Files\n", nFiles);
         for (i=0; i<nFiles; i++) printf("%s \n", fileList[i]);
         unChopData(fileList, nFiles, chan1, chan2, nDrop,
            npix, npix, ATarget1, ATarget2, ASky1, ASky2, &ATargetN, &ASkyN,
            &nCycle, &ierr);
         if (ierr != 0) {
            printf("Failed to process A-Data: error: %i\n", ierr);
            return 0;
         }
/*
       normalize data by dividing by #frames
*/
         normalizeExposure(ATarget1, npix, ATargetN, dit);
         normalizeExposure(ATarget2, npix, ATargetN, dit);
         normalizeExposure(ASky1, npix, ASkyN, dit);
         normalizeExposure(ASky2, npix, ASkyN, dit);
/*
       subtract sky
*/
         for (iPix = 0; iPix< npix; iPix++) {
            ATarget1[iPix] = ATarget1[iPix] - ASky1[iPix];
            ATarget2[iPix] = ATarget2[iPix] - ASky2[iPix];
         }
   /* go back through data and accumlate data in bins */
         printf("Estimating RMS of A-exposures\n");
         binChopData(fileList, nFiles, chan1, chan2, nDrop, 
            npix, npix, nCycle, ABin1, ABin2, &ierr);
         if (ierr != 0) {
            printf("Failed to process ARMS-Data: error: %i\n", ierr);
            return 0;
         }
/*
              normalize 
*/
         for (i=0; i<NBIN; i++) {
            normalizeExposure(ABin1[i], npix, 1., dit);
            normalizeExposure(ABin2[i], npix, 1., dit);
         }
      }
      if (nShutter == 2) {
         printf("Found %i B-photometry Files\n", nFiles);
         for (i=0; i<nFiles; i++) printf("%s \n", fileList[i]);
         unChopData(fileList, nFiles, chan1, chan2, nDrop,
            npix, npix, BTarget1, BTarget2, BSky1, BSky2, &BTargetN, &BSkyN,
            &nCycle, &ierr);
         if (ierr != 0) {
            printf("Failed to process B-Data: error: %i\n", ierr);
            return 0;
         }
         normalizeExposure(BTarget1, npix, BTargetN, dit);
         normalizeExposure(BTarget2, npix, BTargetN, dit);
         normalizeExposure(BSky1, npix, BSkyN, dit);
         normalizeExposure(BSky2, npix, BSkyN, dit);

         for (iPix = 0; iPix< npix; iPix++) {
            BTarget1[iPix] = BTarget1[iPix] - BSky1[iPix];
            BTarget2[iPix] = BTarget2[iPix] - BSky2[iPix];
         }
   /* go back and bin data */
         printf("Estimating RMS of B-exposures\n");
         binChopData(fileList, nFiles, chan1, chan2, nDrop, 
            npix, npix, nCycle, BBin1, BBin2, &ierr);
         if (ierr != 0) {
            printf("Failed to process BRMS-Data: error: %i\n", ierr);
            return 0;
         }
         for (i=0; i<NBIN; i++) {
            normalizeExposure(BBin1[i], npix, 1., dit);
            normalizeExposure(BBin2[i], npix, 1., dit);
         }
      }
      oirKillFileList(fileList, nFiles);
   }   /* loop over shutters */

/* debug dump */
#ifdef DEBUG 
i = fits_create_file(&debugFile, "!photobug.fits", &ierr);
if (ierr !=0) {
   printf("error %i creating debug photo file\n", ierr);
   printf("%s \n",strerror(errno));
   return 1;
}
debugTable = make_imaging_fdata_table_def (debugFile, DetectorDef, 1, &ierr);
debugRow   = make_imaging_fdata_row(debugTable);
memcpy(debugRow->data[0], ATarget1, npix*sizeof(float));
memcpy(debugRow->data[1], ATarget2, npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 1, &ierr);

memcpy(debugRow->data[0], BTarget1, npix*sizeof(float));
memcpy(debugRow->data[1], BTarget2, npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 2, &ierr);

for (ix=0;ix<NBIN;ix++){
memcpy(debugRow->data[0], ABin1[ix], npix*sizeof(float));
memcpy(debugRow->data[1], ABin2[ix], npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 3+ix, &ierr);
}

for (ix=0;ix<NBIN;ix++){
memcpy(debugRow->data[0], BBin1[ix], npix*sizeof(float));
memcpy(debugRow->data[1], BBin2[ix], npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 3+NBIN+ix, &ierr);
}
#endif 
/* debug dump */
/*
      deSky
      y coordinates of sky are cludge for prism; need better algorithm to find background
      set yBottom = 12 for prism, 5 for grism
*/
   if (gris[0] == 'G') yBottom = 6-dSky; else yBottom = 9-dSky;
   if (gris[0] == 'G') yTop = 29+dSky; else yTop = 25+dSky;
   if ((yBottom > 2) && (yTop < 38)) {
      printf("Sky Subtract from %i to %i \n", yBottom, yTop);
      deSky(ATarget1, nx, ny, yBottom, yTop); 
      deSky(ATarget2, nx, ny, yBottom, yTop);
      deSky(BTarget1, nx, ny, yBottom, yTop);
      deSky(BTarget2, nx, ny, yBottom, yTop);
      for (i=0; i<NBIN; i++) {
         deSky(ABin1[i], nx, ny, yBottom, yTop);
         deSky(ABin2[i], nx, ny, yBottom, yTop);
         deSky(BBin1[i], nx, ny, yBottom, yTop);
         deSky(BBin2[i], nx, ny, yBottom, yTop);
      }
   }
   else {
      printf("No Sky Subtraction along slit\n");
      yBottom = 2;
      yTop    = 38;
   }
/*
      get mask array
*/
   oirGetMaskImages(argv[3], maskImages, DetectorDef->nregion, nxList, nyList, &ierr);
   if (ierr != 0) {
      printf("error %i reading mask data from file %s\n", ierr, argv[3]);
      return 1;
   }
/*
      calculate total fluxes 
*/
   for (ix = 0; ix <nx; ix++) {
      for (iy=yBottom; iy <=yTop; iy++) {
         ii = ix + nx*(long)iy;
         ATotal[ix] = ATotal[ix] + ATarget1[ii] + ATarget2[ii];
         BTotal[ix] = BTotal[ix] + BTarget1[ii] + BTarget2[ii];
         xx = ATarget1[ii] * BTarget1[ii];
         if (xx > 0) ABTotal[ix] = ABTotal[ix] + sqrt(xx);
         xx = ATarget2[ii] * BTarget2[ii];
         if (xx > 0) ABTotal[ix] = ABTotal[ix] + sqrt(xx);
      } /* iy */
/*
      calculate rmss from binned totals
*/
      for (i=0; i<NBIN; i++) {
         Ascratch  = 0.;
         Bscratch  = 0.;
         ABscratch = 0.;
         for (iy=yBottom; iy <=yTop; iy++) {
            ii = ix + nx*(long)iy;
            Ascratch = Ascratch + ABin1[i][ii] + ABin2[i][ii];
            Bscratch = Bscratch + BBin1[i][ii] + BBin2[i][ii];
            xx = ABin1[i][ii] * BBin1[i][ii];
            if (xx > 0) ABscratch = ABscratch + sqrt(xx);
            xx = ABin2[i][ii] * BBin2[i][ii];
            if (xx > 0) ABscratch = ABscratch + sqrt(xx);
         } /* iy */
         Ascratch  = Ascratch  - ATotal[ix];
         Bscratch  = Bscratch  - BTotal[ix];
         ABscratch = ABscratch - ABTotal[ix];
         ARTotal[ix]  = ARTotal[ix]  + Ascratch*Ascratch;
         BRTotal[ix]  = BRTotal[ix]  + Bscratch*Bscratch;
         ABRTotal[ix] = ABRTotal[ix] + ABscratch*ABscratch;
      } /* iBin */
      if (ARTotal[ix] > 0)  ARTotal[ix]  = sqrt(ARTotal[ix]/(NBIN*(NBIN-1)));
      if (BRTotal[ix] > 0)  BRTotal[ix]  = sqrt(BRTotal[ix]/(NBIN*(NBIN-1)));
      if (ABRTotal[ix] > 0) ABRTotal[ix] = sqrt(ABRTotal[ix]/(NBIN*(NBIN-1)));
   } /* ix */

/*
      calculate masked fluxes
*/
   for (ix = 0; ix <nx; ix++) {
      for (iy=yBottom; iy <=yTop; iy++) { 
         ii = ix + nx*(long)iy;
         AMask[ix] = AMask[ix] + ATarget1[ii] * maskImages[chan1][ii] +
            ATarget2[ii] * maskImages[chan2][ii];
         BMask[ix] = BMask[ix] + BTarget1[ii] * maskImages[chan1][ii] +
            BTarget2[ii] * maskImages[chan2][ii];
         mm = maskImages[chan1][ii];
         xx = ATarget1[ii]*BTarget1[ii] * mm * mm;
         if (xx > 0) ABMask[ix] = ABMask[ix] + sqrt(xx);
         mm = maskImages[chan2][ii];
         xx = ATarget2[ii]*BTarget2[ii] * mm * mm;
         if (xx > 0) ABMask[ix] = ABMask[ix] + sqrt(xx);
      } /* iy */
/*
      calculate rmss from binned totals
*/
      for (i=0; i<NBIN; i++) {
         Ascratch  = 0.;
         Bscratch  = 0.;
         ABscratch = 0.;
         for (iy=yBottom; iy <=yTop; iy++) {
            ii = ix + nx*(long)iy;
            Ascratch = Ascratch + ABin1[i][ii] * maskImages[chan1][ii] +
               + ABin2[i][ii] * maskImages[chan2][ii];
            Bscratch = Bscratch + BBin1[i][ii] * maskImages[chan1][ii] + 
               BBin2[i][ii] * maskImages[chan2][ii];
            mm = maskImages[chan1][ii];
            xx = ABin1[i][ii]*BBin1[i][ii] * mm * mm;
            if (xx > 0) ABscratch = ABscratch + sqrt(xx);
            mm = maskImages[chan2][ii];
            xx = ABin2[i][ii]*BBin2[i][ii] * mm * mm;
            if (xx > 0) ABscratch = ABscratch + sqrt(xx);
         } /* iy */
         Ascratch  = Ascratch  - AMask[ix];
         Bscratch  = Bscratch  - BMask[ix];
         ABscratch = ABscratch - ABMask[ix];
         ARMask[ix]  = ARMask[ix]  + Ascratch*Ascratch;
         BRMask[ix]  = BRMask[ix]  + Bscratch*Bscratch;
         ABRMask[ix] = ABRMask[ix] + ABscratch*ABscratch;
      } /* iBin */
      if (ARMask[ix] > 0) ARMask[ix] = sqrt(ARMask[ix]/(NBIN*(NBIN-1)));
      if (BRMask[ix] > 0) BRMask[ix] = sqrt(BRMask[ix]/(NBIN*(NBIN-1)));
   } /* ix */
/* debug dump */
#ifdef DEBUG
memcpy(debugRow->data[0], ATarget1, npix*sizeof(float));
memcpy(debugRow->data[1], ATarget2, npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 3+2*NBIN, &ierr);

memcpy(debugRow->data[0], BTarget1, npix*sizeof(float));
memcpy(debugRow->data[1], BTarget2, npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 4+2*NBIN, &ierr);

memcpy(debugRow->data[0], ASky1, npix*sizeof(float));
memcpy(debugRow->data[1], ASky2, npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 5+2*NBIN, &ierr);

memcpy(debugRow->data[0], BSky1, npix*sizeof(float));
memcpy(debugRow->data[1], BSky2, npix*sizeof(float));
i = write_imaging_fdata_table(debugFile, debugTable, debugRow, 6+2*NBIN, &ierr);
i = fits_close_file(debugFile, &ierr);
kill_imaging_fdata_row(debugRow);
/*i = close_imaging_fdata_table(debugFile, debugTable, &ierr);*/
#endif
/*
      prepare output file
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[4]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",argv[4], ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   DetectorDef->nregion = 1;
   DetectorDef->naxis[0][1] = 1;
   iKey = table_find_key(DetectorDef->def, "NREGION");
   table_put_keyword(DetectorDef->def, iKey, (void *)&one, &ierr);

   odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
   }
   outData = make_imaging_fdata_row(odata);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ATotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 2, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 3, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = AMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 4, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 5, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 6, &ierr);
   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ARTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 7, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BRTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 8, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABRTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 9, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ARMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 10, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BRMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 11, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BRMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 12, &ierr);
   i = close_imaging_fdata_table(outFile, odata, &ierr);
   kill_imaging_fdata_row(outData);

   free(ATarget1);
   free(ATarget2);
   free(BTarget1);
   free(BTarget2);
   free(ASky1);
   free(ASky2);
   free(BSky1);
   free(BSky2);
   free(AMask);
   free(BMask);
   free(ABMask);
   free(ATotal);
   free(BTotal);
   free(ABTotal);
   free(ARMask);
   free(BRMask);
   free(ABRMask);
   free(ARTotal);
   free(BRTotal);
   free(ABRTotal);
   for (i=0;i<nRegion;i++) free(maskImages[i]);
   for (i=0; i<NBIN; i++) {
      free(ABin1[i]); 
      free(ABin2[i]); 
      free(BBin1[i]); 
      free(BBin2[i]); 
   }
   return 0;
}
